from importlib import import_module
from types import ModuleType
from typing import Optional

from ..logging import logger


def load_module(plugin_name: str) -> Optional[ModuleType]:
    module_name: str = plugin_name
    while True:
        logger.trace(f"Try to import module by name '{module_name}'")
        try:
            return import_module(module_name)
        except ModuleNotFoundError as exc:
            logger.trace(f"Module '{module_name}' import failed ({exc}).")
            if plugin_name != module_name or "." not in module_name:
                break

            module_name, plugin = module_name.rsplit(".", maxsplit=1)
            logger.trace(f"Consider '{module_name}' as module name and '{plugin}' as plugin name")

    return None



from types import ModuleType
from typing import Callable, Generic, List, Optional, Type, TypeVar

from ..logging import logger
from ..utils import callable_name, check_callable_signature
from .module_loader import load_module
from .object_creator import create_object
from .objects_loader import load_objects


T = TypeVar("T")


class SimpleObjectLoader(Generic[T]):
    def __call__(
        self,
        obj_name: str,
        class_: Type[T],
        *,
        type_check: bool = True,
    ) -> List[Callable]:
        _self: str = f"[{callable_name(class_)} '{obj_name}']"
        module: Optional[ModuleType] = load_module(obj_name)
        if not module:
            logger.info(f"{_self}: Cannot load module by object name '{obj_name}'")
            return []

        any_object_classes: List[Type[T]] = load_objects(module, obj_name=obj_name)

        logger.trace(
            f"{_self}: Objects loaded by object name '{obj_name}': {[obj.__name__ for obj in any_object_classes]}"
        )

        object_classes: List[Type[T]] = [
            obj
            for obj in any_object_classes
            if check_callable_signature(
                obj,
                class_=class_,
                type_check=type_check,
            )
        ]

        logger.trace(
            f"{_self}: classes after signature check: {[p.__name__ for p in object_classes]}"
        )

        objects: List[Callable] = []
        for obj_class in object_classes:
            obj: Optional[Callable] = create_object(obj_class)
            if not obj:
                logger.info(
                    f"{_self}: Object '{callable_name(obj_class)}' can't be created in a simple way. Skip."
                )
                continue
            objects.append(obj)

        logger.trace(
            f"{_self}: objects after classes instantiating: {[callable_name(p) for p in objects]}"
        )

        return objects



from typing import Callable, List, Type

from ..interfaces import ObjectLoader


__all__ = ["load_object_or_raise_error"]


def load_object_or_raise_error(
    object_loader: ObjectLoader[Callable],
    *,
    obj_name: str,
    class_: Type[Callable],
    type_check: bool,
) -> Callable:
    objects: List[Callable] = object_loader(
        obj_name=obj_name,
        class_=class_,
        type_check=type_check,
    )
    if not objects:
        raise ImportError(f"Cannot load object '{obj_name}' by signature '{class_}'")

    return objects[0]



from copy import deepcopy
from typing import Any, Dict, Mapping

from ..interfaces import Plugin
from ..logging import logger
from ..types import SecName, Section
from ..utils import callable_params


def simple_plugin_executor(
    *,
    sections: Mapping[str, Section],
    plugins: Mapping[str, Plugin],
) -> Dict[str, Section]:
    sugar_sections: Dict[SecName, Section] = {}

    section: Section
    section_name: str
    plugin_name: str
    plugin: Plugin

    sections_template: Mapping[str, Section] = deepcopy(sections)

    for section_name, section in sections.items():
        for plugin_name, plugin in plugins.items():
            plugin_params: Dict[str, Any] = callable_params(plugin)

            values: Dict[str, Any] = {
                "section_name": section_name,
                "sections": sections,
            }
            if "section" in plugin_params:
                values["section"] = deepcopy(sugar_sections.get(section_name, section))

            kwargs: Dict[str, Any] = {
                param_name: values[param_name] for param_name in plugin_params
            }

            new_section: Section = plugin(**kwargs)

            if kwargs.get("section") and new_section != kwargs["section"]:
                logger.trace(f"Section '{section_name}' was modified by plugin '{plugin_name}'")

            sugar_sections[section_name] = new_section

            if sections != sections_template:
                raise RuntimeError(
                    f"Plugin '{plugin_name}' modified sections, but it's forbidden."
                )

    _ = [
        sugar_sections.setdefault(section_name, section)
        for section_name, section in sections.items()
    ]

    return sugar_sections



from typing import Dict

from ..types import JSON, SecName, Section


__all__ = ["decode_section", "encode_section"]


def encode_section(content: JSON) -> Dict[SecName, Section]:
    if isinstance(content, list):
        return {str(i): item for i, item in enumerate(content)}

    if isinstance(content, (str, int, float, type(None), bool)):
        return {"": content}

    assert isinstance(content, dict)
    return content


def decode_section(sections: Dict[SecName, Section], *, origin: JSON) -> JSON:
    if isinstance(origin, list):
        return [item for item in sections.values()]

    if isinstance(origin, (str, int, float, type(None), bool)):
        return sections[""]

    assert isinstance(origin, dict)
    return sections



from typing import Dict

from ..interfaces import SugarFileLoader
from ..utils import Final
from .simple_json import loader as simple_json_loader
from .simple_yaml import loader as simple_yaml_loader


loader_by_extension: Final[Dict[str, SugarFileLoader]] = {
    ".yaml": simple_yaml_loader,
    ".yml": simple_yaml_loader,
    ".json": simple_json_loader,
}



import json
import sys
from pathlib import Path

from ..types import JSON


def writer(*, path: Path, content: JSON) -> None:
    with path.open("w", encoding="utf-8") as json_file:
        json.dump(content, json_file)


def to_stdout(content: JSON) -> None:
    json.dump(content, sys.stdout, indent=4)



from inspect import Signature, isclass, isfunction, signature
from types import FunctionType
from typing import Any, Callable, Dict, Optional, Tuple, Type, Union

from ..interfaces import Plugin
from ..logging import logger
from ..utils import callable_name
from .names_and_types import is_builtin


__all__ = [
    "callable_params",
    "callable_signature",
    "check_callable_signature",
    "signature_params",
    "signature_type_check",
]


def callable_signature(obj: Union[Type[Any], FunctionType]) -> Optional[Signature]:
    sign: Optional[Signature] = None

    if isclass(obj) and hasattr(obj, "__call__"):
        logger.trace(f"Object '{callable_name(obj)}' is a class and callable")
        sign = signature(obj.__call__)
    elif isfunction(obj):
        logger.trace(f"Object '{callable_name(obj)}' is a function")
        sign = signature(obj)
    else:
        logger.trace(
            f"Object '{callable_name(obj)}' is neither class with __call__ nor function. skip"
        )

    return sign


def signature_params(
    sign: Signature,
    *,
    ignore_names: Tuple = ("self", "args", "kwargs"),
) -> Dict[str, Any]:
    return {p.name: p.annotation for p in sign.parameters.values() if p.name not in ignore_names}


def callable_params(obj: Callable) -> Dict[str, Any]:
    sign: Optional[Signature]
    if isfunction(obj):
        sign = callable_signature(obj)
    else:
        sign = callable_signature(type(obj))

    if not sign:
        raise ValueError(f"Object '{callable_name(obj)}' is not callable")

    return signature_params(sign)


def signature_type_check(*, check: Signature, origin: Signature) -> bool:
    logger.info("'type_check' option is not implemented yet")
    return True


def check_callable_signature(
    obj: Type[Any],
    class_: Type[Any],
    *,
    type_check: bool = False,
) -> bool:
    obj_sign: Optional[Signature] = callable_signature(obj)
    if is_builtin(obj) or not obj_sign:
        return False

    cls_sign: Signature = signature(class_.__call__)

    obj_params: Dict[str, Any] = signature_params(obj_sign)
    cls_params: Dict[str, Any] = signature_params(cls_sign)

    if class_ is not Plugin:
        if obj_params.keys() != cls_params.keys():
            logger.trace(f"Object '{callable_name(obj)}' has got bad param names")
            return False
    else:
        if not set(obj_params.keys()).intersection(set(cls_params.keys())):
            logger.trace(
                f"Object '{callable_name(obj)}' is not satisfy any of possible 'Plugin' signatures."
            )
            return False

    if type_check:
        return signature_type_check(check=obj_sign, origin=cls_sign)

    return True



from enum import Enum
from pathlib import Path
from typing import List

import typer

from .file_loader import loader_by_extension
from .file_writer import stdout_writer_by_extension, writer_by_extension
from .interfaces import (
    FinalFileWriter,
    ObjectLoader,
    PluginExecutor,
    SectionDecoder,
    SectionEncoder,
    SugarFileLoader,
)
from .logging import LogLevel, logger, set_log_level
from .object_loader import SimpleObjectLoader
from .plugin_executor import simple_plugin_executor
from .sugarator import sugarate
from .utils import Final, decode_section, encode_section, load_object_or_raise_error


STDOUT: Final[str] = "stdout"
AUTO: Final[str] = "auto"

app = typer.Typer()


@app.command()
def main(
    file: str,
    plugin: List[str] = typer.Option(...),
    *,
    output: str = STDOUT,
    obj_loader: str = AUTO,
    file_loader: str = AUTO,
    file_writer: str = AUTO,
    sec_encoder: str = AUTO,
    sec_decoder: str = AUTO,
    plug_executor: str = AUTO,
    log_level: Enum("", {level.name: level.name for level in LogLevel}) = LogLevel.INFO.name,  # type: ignore
) -> None:
    set_log_level(LogLevel[log_level.value])
    file_path: Path = Path(file)
    if not file_path.is_file():
        logger.warning(f"File '{file}' does not exist or is not a file")
        return

    object_loader: ObjectLoader = SimpleObjectLoader()
    if obj_loader != AUTO:
        object_loader = load_object_or_raise_error(
            object_loader,
            obj_name=obj_loader,
            class_=ObjectLoader,
            type_check=False,
        )

    sugar_file_loader: SugarFileLoader
    if file_loader != AUTO:
        sugar_file_loader = load_object_or_raise_error(
            object_loader,
            obj_name=file_loader,
            class_=SugarFileLoader,
            type_check=False,
        )
    else:
        ext: str = file_path.suffix
        if ext not in loader_by_extension:
            logger.warning(f"Cannot find plugin loader by extension '{ext}'")
            return

        sugar_file_loader = loader_by_extension[ext]

    sugar_file_writer: FinalFileWriter

    output_path: Path = file_path
    if file_writer != AUTO:
        sugar_file_writer = load_object_or_raise_error(
            object_loader,
            obj_name=file_writer,
            class_=FinalFileWriter,
            type_check=False,
        )
    else:
        write_obj_by_ext = stdout_writer_by_extension
        out_ext: str = file_path.suffix

        if output != STDOUT:
            output_path = Path(output)
            out_ext = output_path.suffix
            write_obj_by_ext = writer_by_extension

        if out_ext not in write_obj_by_ext:
            logger.warning(f"cannot find plugin writer by extension '{out_ext}'")
            return

        sugar_file_writer = write_obj_by_ext[out_ext]

    section_encoder: SectionEncoder = encode_section
    section_decoder: SectionDecoder = decode_section

    if sec_encoder != AUTO:
        section_encoder = load_object_or_raise_error(
            object_loader,
            obj_name=sec_encoder,
            class_=SectionEncoder,
            type_check=False,
        )

    if sec_decoder != AUTO:
        section_decoder = load_object_or_raise_error(
            object_loader,
            obj_name=sec_decoder,
            class_=SectionDecoder,
            type_check=False,
        )

    plugin_executor: PluginExecutor = simple_plugin_executor
    if plug_executor != AUTO:
        plugin_executor = load_object_or_raise_error(
            object_loader,
            obj_name=plug_executor,
            class_=PluginExecutor,
            type_check=False,
        )

    sugarate(
        plugin_name_list=plugin,
        object_loader=object_loader,
        sugar_file_path=file_path,
        sugar_file_loader=sugar_file_loader,
        final_file_path=output_path,
        final_file_writer=sugar_file_writer,
        section_encoder=section_encoder,
        section_decoder=section_decoder,
        plugin_executor=plugin_executor,
        type_check=False,
    )


if __name__ == "__main__":
    app()



from typing import Any, Dict, List, Union


__all__ = ["JSON", "SecName", "Section"]


JSON = Union[
    Dict[str, "JSON"],
    List["JSON"],
    str,
    int,
    float,
    bool,
    None,
]

# Section = Union[JSON, BaseTypedDict, Mapping[str, Any], Dict[str, Any]]
Section = Any  # In fact, section is anything at all
SecName = str



import json
from pathlib import Path

from ..logging import logger
from ..types import JSON


def loader(path: Path) -> JSON:
    with path.open(encoding="utf-8") as load_file:
        return json.load(load_file)



from pathlib import Path
from typing import Dict, List

from .interfaces import (
    FinalFileWriter,
    ObjectLoader,
    Plugin,
    PluginExecutor,
    SectionDecoder,
    SectionEncoder,
    SugarFileLoader,
)
from .logging import logger
from .types import JSON, SecName, Section
from .utils import callable_name


__all__ = ["sugarate"]


def sugarate(
    *,
    plugin_name_list: List[str],
    object_loader: ObjectLoader,
    sugar_file_path: Path,
    sugar_file_loader: SugarFileLoader,
    final_file_path: Path,
    final_file_writer: FinalFileWriter,
    section_encoder: SectionEncoder,
    section_decoder: SectionDecoder,
    plugin_executor: PluginExecutor,
    type_check: bool = True,
) -> None:
    logger.trace(f"Plugin list: {plugin_name_list}")

    plugin_name: str
    plugins: Dict[str, Plugin] = {}
    for plugin_name in plugin_name_list:
        plugin_list: List[Plugin] = object_loader(
            plugin_name,
            class_=Plugin,
            type_check=type_check,
        )
        if not plugin_list:
            logger.warning(f"Unable to load plugin '{plugin_name}', skip")
            continue

        for plugin in plugin_list:
            plugins[f"{plugin_name}.{callable_name(plugin)}"] = plugin

    if not plugins:
        logger.warning("No one plugin was loaded, returning...")
        return

    logger.trace(f"Plugins loaded: {list(plugins.keys())}")

    logger.trace(f"Loading file '{sugar_file_path.name}'")
    content: JSON = sugar_file_loader(sugar_file_path)
    sections: Dict[SecName, Section] = section_encoder(content)

    logger.trace(f"File '{sugar_file_path.name}' successfully loaded")

    sugar_sections: Dict[str, Section] = plugin_executor(
        sections=sections,
        plugins=plugins,
    )

    logger.trace("All plugins were successfully applied. Decoding sections...")

    output_content: JSON = section_decoder(sugar_sections, origin=content)

    logger.trace("Sections were successfully decoded. Writing content to file...")
    final_file_writer(path=final_file_path, content=output_content)



import sys

from ..logging import logger


try:
    import yaml  # type: ignore
except ModuleNotFoundError:
    logger.error("Cannot import module named 'yaml'. Try install 'pyyaml' first and try again")
    raise

from pathlib import Path

from ..types import JSON


def writer(*, path: Path, content: JSON) -> None:
    with path.open("w", encoding="utf-8") as yaml_file:
        yaml.dump(
            content,
            yaml_file,
            width=1024,
        )


def to_stdout(content: JSON) -> None:
    yaml.dump(content, sys.stdout, width=1024)



from ..logging import logger


try:
    import yaml  # type: ignore
except ModuleNotFoundError:
    logger.error("Cannot import module named 'yaml'. Try install 'pyyaml' first and try again")
    raise
from pathlib import Path

from sugaru.utils import Final
from ..types import JSON


class Ref(yaml.YAMLObject):
    yaml_tag = "!reference"
    yaml_loader = yaml.SafeLoader

    def __init__(self, values: yaml.nodes.CollectionNode) -> None:
        self._values: Final[yaml.nodes.CollectionNode] = values

    @classmethod
    def from_yaml(cls, loader, node):
        return cls(loader.construct_sequence(node))

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_sequence(
            cls.yaml_tag,
            data._values,
        )


def loader(path: Path) -> JSON:
    with path.open(encoding="utf-8") as load_file:
        return yaml.safe_load(load_file)



from typing import Dict

from ..interfaces import FinalFileWriter
from ..utils import Final
from .simple_json import to_stdout as json_to_stdout
from .simple_json import writer as simple_json_writer
from .simple_yaml import to_stdout as yaml_to_stdout
from .simple_yaml import writer as simple_yaml_writer


writer_by_extension: Final[Dict[str, FinalFileWriter]] = {
    ".yaml": simple_yaml_writer,
    ".yml": simple_yaml_writer,
    ".json": simple_json_writer,
}

stdout_writer_by_extension: Final[Dict[str, FinalFileWriter]] = {
    ".yaml": lambda path, content: yaml_to_stdout(content),
    ".yml": lambda path, content: yaml_to_stdout(content),
    ".json": lambda path, content: json_to_stdout(content),
}



import sys
from enum import Enum
from unittest.mock import Mock


try:
    from loguru import logger
except ModuleNotFoundError:
    logger = Mock()
    logger.trace = print
    logger.debug = print
    logger.info = print
    logger.warning = print
    logger.error = print


class LogLevel(Enum):
    TRACE = 1
    DEBUG = 2
    INFO = 3
    WARNING = 4
    ERROR = 5


def set_log_level(level: LogLevel) -> None:
    logger.remove()
    logger.add(sys.stdout, level=level.name)
    if isinstance(logger, Mock):
        for log_level in LogLevel:
            if log_level.value < level.value:
                setattr(logger, log_level.name.lower(), lambda *args, **kwargs: None)
            else:
                setattr(logger, log_level.name.lower(), print)



import sys
from pathlib import Path
from typing import Dict, List, Mapping, Type, TypeVar


if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

from .types import JSON, SecName, Section


T = TypeVar("T")

__all__ = [
    "FinalFileWriter",
    "ObjectLoader",
    "Plugin",
    "PluginExecutor",
    "SectionDecoder",
    "SectionEncoder",
    "SugarFileLoader",
]


class Plugin(Protocol):
    def __call__(
        self,
        *,
        section: Section,
        section_name: str,
        sections: Mapping[SecName, Section],
    ) -> Section:
        ...


class SugarFileLoader(Protocol):
    def __call__(self, path: Path) -> JSON:
        ...


class FinalFileWriter(Protocol):
    def __call__(self, *, path: Path, content: JSON) -> None:
        ...


class ObjectLoader(Protocol[T]):
    def __call__(
        self,
        obj_name: str,
        class_: Type[T],
        *,
        type_check: bool = True,
    ) -> List[T]:
        ...


class SectionEncoder(Protocol):
    def __call__(self, content: JSON) -> Dict[SecName, Section]:
        ...


class SectionDecoder(Protocol):
    def __call__(self, sections: Dict[SecName, Section], *, origin: JSON) -> JSON:
        ...


class PluginExecutor(Protocol):
    def __call__(
        self,
        *,
        sections: Mapping[str, Section],
        plugins: Mapping[str, Plugin],
    ) -> Dict[str, Section]:
        ...



from inspect import getmembers, isclass, isfunction
from types import ModuleType
from typing import List, Type, TypeVar

from ..logging import logger
from ..utils import is_builtin


T = TypeVar("T")


def load_objects(
    module: ModuleType,
    *,
    obj_name: str,
) -> List[Type[T]]:
    if module.__name__ == obj_name:
        logger.trace(f"Loading all module '{obj_name}' objects")
        return [
            v
            for _, v in getmembers(
                module,
                predicate=lambda obj: (isclass(obj) or isfunction(obj)) and not is_builtin(obj),
            )
        ]
    else:
        module_name, obj_name = obj_name.rsplit(".", maxsplit=1)
        logger.trace(f"Loading single object '{obj_name}' in module '{module_name}'")
        try:
            return [getattr(module, obj_name)]
        except AttributeError:
            logger.warning(f"No object called '{obj_name}' found in module {module.__name__}")
            return []



from .file_loader import *
from .file_writer import *
from .interfaces import *
from .object_loader import *
from .plugin_executor import *
from .sugarator import *
from .types import *
from .utils import *



import sys

from .encode_decode import *
from .names_and_types import *
from .object_loading import *
from .signature import *


if sys.version_info < (3, 8):
    from typing_extensions import Final
else:
    from typing import Final  # noqa



from inspect import isclass, isfunction
from typing import Callable, Optional, Type, TypeVar


T = TypeVar("T")


def create_object(plugin_class: Type[T]) -> Optional[Callable]:
    if isclass(plugin_class):
        try:
            return plugin_class()
        except TypeError:
            return None

    elif isfunction(plugin_class):
        return plugin_class

    return None



import builtins
import typing
from types import FunctionType
from typing import Any, Set, Type, Union

import typing_extensions


__all__ = ["BUILTIN_TYPES", "callable_name", "is_builtin"]

BUILTIN_TYPES: Set[str] = set(dir(builtins) + dir(typing) + dir(typing_extensions))


def is_builtin(_type: Union[Type, FunctionType]) -> bool:
    return _type.__name__ in BUILTIN_TYPES


def callable_name(obj: Any) -> str:
    if hasattr(obj, "__name__"):
        return obj.__name__

    return type(obj).__name__
